/*
 *Copyright (C) 1994-2000 The XFree86 Project, Inc. All Rights Reserved.
 *
 *Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 *"Software"), to deal in the Software without restriction, including
 *without limitation the rights to use, copy, modify, merge, publish,
 *distribute, sublicense, and/or sell copies of the Software, and to
 *permit persons to whom the Software is furnished to do so, subject to
 *the following conditions:
 *
 *The above copyright notice and this permission notice shall be
 *included in all copies or substantial portions of the Software.
 *
 *THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 *EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 *MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 *NONINFRINGEMENT. IN NO EVENT SHALL THE XFREE86 PROJECT BE LIABLE FOR
 *ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
 *CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 *WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 *
 *Except as contained in this notice, the name of the XFree86 Project
 *shall not be used in advertising or otherwise to promote the sale, use
 *or other dealings in this Software without prior written authorization
 *from the XFree86 Project.
 *
 * Authors:	Harold L Hunt II
 *
 * Modifier:	MATSUZAKI Kensuke
 */


/* Standard library headers */
#include <assert.h>
#include <stdio.h>
#include <stdlib.h>


/* X headers */
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xos.h>
#include <X11/Xatom.h>
#include <X11/keysym.h>
#include <X11/Xlocale.h>

/* Fixups to prevent collisions between Windows and X headers */
#undef MINSHORT
#undef MAXSHORT

/* Flags for Windows header options */
#define NONAMELESSUNION
#ifndef WIN32_LEAN_AND_MEAN
#define WIN32_LEAN_AND_MEAN
#endif

/* Windows headers */
#include <windows.h>
#include <windowsx.h>
#include <mmsystem.h>


/* Application constants */
#define WINDOW_CLASS		"xwinclip"
#define WINDOW_TITLE		"xwinclip"


/*
 * Global variables
 */

Display			*pDisplay;
Window			iWindow;
Atom			atomClipboard, atomClipboardManager;
Atom			atomReturnType;
Atom			atomLocalProperty;
Atom			atomUTF8String;
Atom			atomCompoundText;
HWND			hwnd;
HGLOBAL         hGlobal;
HWND            hwndNextViewer;
BOOL			initialized = FALSE;
BOOL            isOwner = FALSE;
/*
 * Function prototypes
 */

HWND
CreateMessagingWindow ();

void
FlushWindowsMessageQueue (HWND hwnd);

LRESULT CALLBACK
WindowProc (HWND hwnd, UINT message, 
			WPARAM wParam, LPARAM lParam);

void
DOStoUNIX (char *pszData, int iLength);

void
UNIXtoDOS (unsigned char **ppszData, int iLength);

void
OnSelectionClear(const XEvent* event);

void
OnSelectionRequest(const XEvent* event);

void
OnSelectionNotify(const XEvent* event);

int
OnExit(Display* dsp);

void
QueryClipboard();

void
PrintWindowsLastError();

int
main (int argc, char *argv[])
{
	XEvent		event;
	XSelectionEvent	eventSelection;
#if 1
	char			*pszAtom = NULL;
	char			*pszProperty = NULL;
#endif
	int			iReturn;
	int           nextConversionTime = 0;
	/* Create Windows messaging window */
	hwnd = CreateMessagingWindow ();

	if (!setlocale(LC_ALL, ""))
	{
		fprintf (stderr, "setlocale() error\n");
		exit (1);
	}
	if (XSupportsLocale() == False)
	{
		fprintf (stderr, "Locale not supported by X\n");
		exit (1);
	}
	/* Open the X display */
	pDisplay = XOpenDisplay (NULL);
	if (pDisplay == NULL)
    {
		printf ("Could not open display\n");
		exit (1);
    }

	/* Create a messaging window */
	iWindow = XCreateSimpleWindow (pDisplay,
								   RootWindow (pDisplay, 0),
								   1, 1,
								   50, 50,
								   0,
								   BlackPixel (pDisplay, 0),
								   BlackPixel (pDisplay, 0));
	if (iWindow == 0)
    {
		printf ("Could not create a window\n");
		exit (1);
    }

#if 1
	/* Display the messaging window, as a test */
	XMapWindow (pDisplay, iWindow);
#endif
  
	/* What does this do? */
	XFlush (pDisplay);

	/* Select event types to watch */
	XSelectInput (pDisplay, iWindow,
				  ExposureMask|StructureNotifyMask);
  
	/* Create an atom for CLIPBOARD_MANAGER */
	atomClipboardManager = XInternAtom (pDisplay, "CLIPBOARD_MANAGER", False);
	if (atomClipboardManager == None)
    {
		printf ("Could not create CLIPBOARD_MANAGER atom\n");
		exit (1);
    }

	/* Assert ownership of CLIPBOARD_MANAGER */
	iReturn = XSetSelectionOwner (pDisplay, atomClipboardManager,
								  iWindow, CurrentTime);
	if (iReturn == BadAtom || iReturn == BadWindow)
    {
		printf ("Could not set CLIPBOARD_MANAGER owner\n");
		exit (1);
    }

	/* Create an atom for CLIPBOARD */
	atomClipboard = XInternAtom (pDisplay, "CLIPBOARD", False);
	if (atomClipboard == None)
    {
		printf ("Could not create CLIPBOARD atom\n");
		exit (1);
    }

	/* Assert ownership of CLIPBOARD */
	iReturn = XSetSelectionOwner (pDisplay, atomClipboard,
								  iWindow, CurrentTime);
	if (iReturn == BadAtom || iReturn == BadWindow)
    {
		printf ("Could not set CLIPBOARD owner\n");
		exit (1);
    }

#if 0
	/* Assert ownership of PRIMARY */
	iReturn = XSetSelectionOwner (pDisplay, XA_PRIMARY,
								  iWindow, CurrentTime);
	if (iReturn == BadAtom || iReturn == BadWindow)
    {
		printf ("Could not set PRIMARY owner\n");
		exit (1);
    }
#endif
	/* Local property to hold pasted data */
	atomLocalProperty = XInternAtom (pDisplay, "CYGX_CUT_BUFFER", False);
	if (atomLocalProperty == None)
    {
		printf ("Could not create CYGX_CUT_BUFFER atom\n");
		exit (1);
    }

	/* Create an atom for UTF8_STRING */
	atomUTF8String = XInternAtom (pDisplay, "UTF8_STRING", False);
	if (atomUTF8String == None)
    {
		printf ("Could not create UTF8_STRING atom\n");
		exit (1);
    }

	/* Create an atom for COMPOUND_TEXT */
	atomCompoundText = XInternAtom (pDisplay, "COMPOUND_TEXT", False);
	if (atomCompoundText == None)
	{
		printf ("Could not create COMPOUND_TEXT atom\n");
		exit (1);
	}

	XSetIOErrorHandler(OnExit);
	initialized = TRUE;
	/* Loop for X events */
	while (1)
	{
		/* Process Windows messages */
		FlushWindowsMessageQueue (hwnd);

		if(!XPending(pDisplay)){
			if(!isOwner){
				QueryClipboard(&event);
			}
			Sleep(500);
			continue;
		}
		/* Block until the next X event */
		XNextEvent (pDisplay, &event);
	  
		/* Branch on the event type */
		switch (event.type)
		{
		case SelectionClear:
			OnSelectionClear(&event);
			break;
		  
			/*
			 * SelectionRequest
			 */
		  
		case SelectionRequest:
			OnSelectionRequest(&event);
			break;
		  
			/*
			 * SelectionNotify
			 */ 
		  
		case SelectionNotify:
			OnSelectionNotify(&event);
			break;

		case DestroyNotify:
			printf("DestroyNotify\n");
			break;

		case ConfigureNotify:
			printf("ConfigureNotify\n");
			break;
		default:
			break;
		}
    }
  
#if 1
	/* Pause for 5 seconds before closing */
	sleep (5);
#endif

	return 0;
}


/*
 * Change \n to \r\n.  Reallocate string if necessary.
 */

void
UNIXtoDOS (unsigned char **ppszData, int iLength)
{
	int			iNewlineCount = 0;
	unsigned char		*pszSrc = *ppszData;
	unsigned char		*pszEnd = pszSrc + iLength;
	unsigned char		*pszDest = NULL, *pszDestBegin = NULL;

#if 0
	printf ("UNIXtoDOS () - Original data:\n%s\n", *ppszData);
#endif

	/* Count \n characters without leading \r */
	while (pszSrc < pszEnd)
    {
		/* Skip ahead two character if found set of \r\n */
		if (*pszSrc == '\r' && pszSrc + 1 < pszEnd && *(pszSrc + 1) == '\n')
		{
			pszSrc += 2;
			continue;
		} 

		/* Increment the count if found naked \n */
		if (*pszSrc == '\n')
		{
			iNewlineCount++;
		}

		pszSrc++;
    }
  
	/* Return if no naked \n's */
	if (iNewlineCount == 0)
		return;

	/* Allocate a new string */
	pszDestBegin = pszDest = malloc (iLength + iNewlineCount + 1);

	/* Set source pointer to beginning of data string */
	pszSrc = *ppszData;

	/* Loop through all characters in source string */
	while (pszSrc < pszEnd)
    {
		/* Copy line endings that are already valid */
		if (*pszSrc == '\r' && pszSrc + 1 < pszEnd && *(pszSrc + 1) == '\n')
		{
			*pszDest = *pszSrc;
			*(pszDest + 1) = *(pszSrc + 1);
			pszDest += 2;
			pszSrc += 2;
			continue;
		}

		/* Add \r to naked \n's */
		if (*pszSrc == '\n')
		{
			*pszDest = '\r';
			*(pszDest + 1) = *pszSrc;
			pszDest += 2;
			pszSrc += 1;
			continue;
		}

		/* Copy normal characters */
		*pszDest = *pszSrc;
		pszSrc++;
		pszDest++;
    }

	/* Put terminating null at end of new string */
	*pszDest = '\0';

	/* Swap string pointers */
	free (*ppszData);
	*ppszData = pszDestBegin;

#if 0
	printf ("UNIXtoDOS () - Final string:\n%s\n", pszDestBegin);
#endif
}


/*
 * This was heavily inspired by, if not down right stolen from, Cygwin's
 * winsup/cygwin/fhandler.cc/fhandler_base::read ()
 */

void
DOStoUNIX (char *pszSrc, int iLength)
{
	char			*pszDest = pszSrc;
	char			*pszEnd = pszSrc + iLength;

	/* Loop until the last character */
	while (pszSrc < pszEnd)
    {
		/* Copy the current source character to current destination character */
		*pszDest = *pszSrc;

		/* Advance to the next source character */
		pszSrc++;

		/* Don't advance the destination character if we need to drop an \r */
		if (*pszDest != '\r' || *pszSrc != '\n')
			pszDest++;
    }

	/* Move the terminating null */
	*pszDest = '\0';
}


void
FlushWindowsMessageQueue (HWND hwnd)
{
	MSG			msg;

	/* Flush the messaging window queue */
	while (PeekMessage (&msg, hwnd, 0, 0, PM_REMOVE))
    {
		DispatchMessage (&msg);
    } 
}


LRESULT CALLBACK
WindowProc (HWND hwnd, UINT message, 
			WPARAM wParam, LPARAM lParam)
{
	/* Branch on message type */
	switch (message)
    {
    case WM_CREATE:
#if 1
		printf ("WindowProc - WM_CREATE\n");
#endif
		hwndNextViewer = SetClipboardViewer(hwnd);
		return 0;
    case WM_DESTROY:
#if 1
		printf ("WindowProc - WM_DESTROY\n");
#endif
		return 0;
	  
	case WM_CHANGECBCHAIN:
#if 1
		printf ("WindowProc - WM_CHANGECBCHAIN\n");
#endif
		if(hwndNextViewer == (HWND)wParam){
			hwndNextViewer = (HWND)lParam;
		}else{
			SendMessage(hwndNextViewer, WM_CHANGECBCHAIN, wParam, lParam);
		}
		return 0;
	  
	case WM_DRAWCLIPBOARD:
#if 1
		printf ("WindowProc - WM_DRAWCLIPBOARD\n");
#endif
		if (hwndNextViewer){
			SendMessage (hwndNextViewer, message, wParam, lParam) ;
		}
		if(GetClipboardOwner() != hwnd){
			int iReturn;
			int format;
			if(!initialized){
				return 0;
			}
			if(!OpenClipboard(hwnd)){
				PrintWindowsLastError();
				exit(1);
			}
			format = 0;
			do{
				format = EnumClipboardFormats(format);
				if(format == CF_TEXT){
					break;
				}
			}while(format != 0);
			if(!CloseClipboard()){
				PrintWindowsLastError();
				exit(1);
			}
			
			printf(":");
			if(format == CF_TEXT){
#if 1
				printf ("  Windows Clipboard Updated.\n");
#endif
				/* Assert ownership of PRIMARY */
				iReturn = XSetSelectionOwner (pDisplay, XA_PRIMARY,
											  iWindow, CurrentTime);
				if (iReturn == BadAtom || iReturn == BadWindow)
				{
					printf ("Could not set PRIMARY owner\n");
				}
				isOwner = TRUE;
			}
		}
		return 0;
	  
    default:
		break;
    }

	/* Let Windows perform default processing for unhandled messages */
	return DefWindowProc (hwnd, message, wParam, lParam);
}

void
QueryClipboard()
{
	int			iReturn;
	/* Request the lost selection contents */
	iReturn = XConvertSelection (pDisplay,
								 XA_PRIMARY,
								 atomCompoundText,//XA_STRING,
								 atomLocalProperty,
								 iWindow,
								 CurrentTime);
	if (iReturn == BadAtom || iReturn == BadWindow)
	{
		printf ("QueryClipboard - XConvertSelection () failed\n");
		exit (1);
	}
}

void
OnSelectionClear(const XEvent* event)
{
	isOwner = FALSE;
}

void
OnSelectionRequest(const XEvent* event)
{
	int			iReturn;
	XSelectionEvent	eventSelection;
	char			*pszGlobalData;
#if 1
	printf ("SelectionRequest - target %d\n",
			event->xselectionrequest.target);
	printf ("target name is %s\n",
			XGetAtomName(pDisplay, event->xselectionrequest.target));
#endif
	/* Abort if invalid target type */
	if (event->xselectionrequest.target == XA_STRING
		&& event->xselectionrequest.target == atomUTF8String){
		/* Access the clipboard */
		if (!OpenClipboard (hwnd))
		{
			printf ("OpenClipboard () failed\n");
			PrintWindowsLastError();
			exit (1);
		}
		
		/* Get a pointer to the clipboard text */
		hGlobal = GetClipboardData (CF_TEXT);
		if (!hGlobal)
		{
			printf ("GetClipboardData () failed\n");
			PrintWindowsLastError();
			exit (1);
		}
		pszGlobalData = (char *) GlobalLock (hGlobal);
  
		/* Convert DOS string to UNIX string */
		DOStoUNIX (pszGlobalData, strlen (pszGlobalData));
		
		/* Copy the clipboard text to the requesting window */
		iReturn = XChangeProperty (pDisplay,
								   event->xselectionrequest.requestor,
								   event->xselectionrequest.property,
								   event->xselectionrequest.target,
								   8,
								   PropModeReplace,
								   pszGlobalData,
								   strlen (pszGlobalData));
		if (iReturn == BadAlloc || iReturn == BadAtom
			|| iReturn == BadMatch || iReturn == BadValue
			|| iReturn == BadWindow)
		{
			printf ("SelectionRequest - XChangeProperty failed: %d\n",
					iReturn);
			exit (1);
		}
		
		/* Release the clipboard data */
		GlobalUnlock (hGlobal);
		pszGlobalData = NULL;
		if(!CloseClipboard ()){
			PrintWindowsLastError();
			exit(1);
		}
		
		/* Setup selection notify event */
		eventSelection.type = SelectionNotify;
		eventSelection.send_event = True;
		eventSelection.display = pDisplay;
		eventSelection.requestor = event->xselectionrequest.requestor;
		eventSelection.selection = event->xselectionrequest.selection;
		eventSelection.target = event->xselectionrequest.target;
		eventSelection.property = event->xselectionrequest.property;
		eventSelection.time = event->xselectionrequest.time;
		
		/* Notify the requesting window that the operation has completed */
		iReturn = XSendEvent (pDisplay,
							  eventSelection.requestor,
							  False,
							  0L,
							  (XEvent *) &eventSelection);
		if (iReturn == BadValue || iReturn == BadWindow)
		{
			printf ("XSendEvent () failed\n");
			exit (1);
		}
	}
	else if (event->xselectionrequest.target == atomCompoundText)
	{
		char *l[2];
		XTextProperty p;
		
		/* Access the clipboard */
		if (!OpenClipboard (hwnd))
		{
			printf ("OpenClipboard () failed\n");
			PrintWindowsLastError();
			exit (1);
		}
		
		/* Get a pointer to the clipboard text */
		hGlobal = GetClipboardData (CF_TEXT);
		if (!hGlobal)
		{
			printf ("GetClipboardData () failed\n");
			PrintWindowsLastError();
			exit (1);
		}
		pszGlobalData = (char *) GlobalLock (hGlobal);
  
		/* Convert DOS string to UNIX string */
		DOStoUNIX (pszGlobalData, strlen (pszGlobalData));
		
		l[0] = pszGlobalData;
		l[1] = NULL;
		p.value = NULL;
		
		/* Copy the clipboard text to the requesting window */
		iReturn = XmbTextListToTextProperty (pDisplay,
											 l,
											 1,
											 XCompoundTextStyle,
											 &p);
		printf ("%s\n", XGetAtomName(pDisplay, p.encoding));
		
		if (iReturn == XNoMemory || iReturn == XLocaleNotSupported)
		{
			printf ("SelectionRequest - XmbTextListToTextProperty failed: %d\n",
					iReturn);
			exit(0);
		}else{
			printf ("SelectionRequest - XmbTextListToTextProperty : %d\n",
					iReturn);
		}
		iReturn = XChangeProperty (pDisplay,
								   event->xselectionrequest.requestor,
								   event->xselectionrequest.property,
								   event->xselectionrequest.target,
								   8,
								   PropModeReplace,
								   p.value,
								   p.nitems);
		if (iReturn == BadAlloc || iReturn == BadAtom
			|| iReturn == BadMatch || iReturn == BadValue
			|| iReturn == BadWindow)
		{
			printf ("SelectionRequest - XChangeProperty failed: %d\n",
					iReturn);
			exit (1);
		}
		
		/* Release the clipboard data */
		GlobalUnlock (hGlobal);
		pszGlobalData = NULL;
		if(!CloseClipboard ()){
			PrintWindowsLastError();
			exit(1);
		}
		
		/* Setup selection notify event */
		eventSelection.type = SelectionNotify;
		eventSelection.send_event = True;
		eventSelection.display = pDisplay;
		eventSelection.requestor = event->xselectionrequest.requestor;
		eventSelection.selection = event->xselectionrequest.selection;
		eventSelection.target = event->xselectionrequest.target;
		eventSelection.property = event->xselectionrequest.property;
		eventSelection.time = event->xselectionrequest.time;
				
		/* Notify the requesting window that the operation has completed */
		iReturn = XSendEvent (pDisplay,
							  eventSelection.requestor,
							  False,
							  0L,
							  (XEvent *) &eventSelection);
		if (iReturn == BadValue || iReturn == BadWindow)
		{
			printf ("XSendEvent () failed\n");
			exit (1);
		}
		//if (p.value){
		//	XFree (p.value);
		//}
	}
	else
	{
		/* Setup selection notify event */
		eventSelection.type = SelectionNotify;
		eventSelection.send_event = True;
		eventSelection.display = pDisplay;
		eventSelection.requestor = event->xselectionrequest.requestor;
		eventSelection.selection = event->xselectionrequest.selection;
		eventSelection.target = event->xselectionrequest.target;
		eventSelection.property = None;
		eventSelection.time = event->xselectionrequest.time;
	  
		/* Notify the requesting window that the operation is complete */
		iReturn = XSendEvent (pDisplay,
							  eventSelection.requestor,
							  False,
							  0L,
							  (XEvent *) &eventSelection);
		if (iReturn == BadValue || iReturn == BadWindow)
		{
			printf ("XSendEvent () failed\n");
			exit (1);
		}
		return;
	}
}

void
OnSelectionNotify(const XEvent* event)
{
	int			iReturn;
	int			iReturnFormat;
	unsigned long		ulReturnItems, ulReturnBytesLeft;
	unsigned char		*pszReturnData = NULL;
	char			*pszGlobalData;
	static char* pszPrevClipboardData = NULL;

#if 1
	printf ("SelectionNotify\n");
#endif
	if((event->xselection.target == XA_STRING)
	   ||(event->xselection.target == atomUTF8String))
	{
#if 1
		printf ("SelectionNotify XA_STRING\n");
#endif
		if(event->xselection.property == None){
#if 0
			printf("Selection XA_STRING is not Avaliable.\n");
#endif
			return;
		}
		/* Retrieve the size of the stored data */
		iReturn = XGetWindowProperty (pDisplay,
									  iWindow,
									  atomLocalProperty,
									  0,
									  0, /* Don't get data, just get size */
									  False,
									  AnyPropertyType,
									  &atomReturnType,
									  &iReturnFormat,
									  &ulReturnItems,
									  &ulReturnBytesLeft,
									  &pszReturnData);
		if (iReturn != Success)
		{
			printf ("SelectionNotify - XGetWindowProperty () failed\n");
			exit (1);
		}
			  
#if 0
		printf ("SelectionNotify - returned data %d left %d\n",
				ulReturnItems, ulReturnBytesLeft);
#endif
			  
		/* Request the selection data */
		iReturn = XGetWindowProperty (pDisplay,
									  iWindow,
									  atomLocalProperty,
									  0,
									  ulReturnBytesLeft,
									  False,
									  AnyPropertyType,
									  &atomReturnType,
									  &iReturnFormat,
									  &ulReturnItems,
									  &ulReturnBytesLeft,
									  &pszReturnData);
		if (iReturn != Success)
		{
			printf ("SelectionNotify - XGetWindowProperty () failed\n");
			exit (1);
		}  
#if 0
		printf ("SelectionNotify - returned data %d left %d\n",
				ulReturnItems, ulReturnBytesLeft);
#endif		  
		/* Convert the X clipboard string to DOS format */
		UNIXtoDOS (&pszReturnData, strlen (pszReturnData));
	}
	else if(event->xselection.target == atomCompoundText)
	{
		XTextProperty p;
		char **l;
		int n;
		unsigned long left;
#if 1
		printf ("SelectionNotify Compound Text\n");
#endif
		if(event->xselection.property == None){
#if 1
			printf("Selection Compound Text is not Avaliable. Try XA_STRING\n");
#endif
			iReturn = XConvertSelection (pDisplay,
										 XA_PRIMARY,
										 XA_STRING,
										 atomLocalProperty,
										 iWindow,
										 CurrentTime);
			if (iReturn == BadAtom || iReturn == BadWindow)
			{
				printf ("SelectionNotify - XConvertSelection () failed\n");
				exit (1);
			}
			return;
		}
		iReturn = XGetWindowProperty(pDisplay,
									 iWindow,
									 atomLocalProperty,
									 0,
									 0,
									 False,
									 atomCompoundText,
									 &p.encoding,
									 &p.format,
									 &p.nitems,
									 &left,
									 &p.value);
		if (iReturn != Success)
		{
			printf ("SelectionNotify - XGetWindowProperty ()1 failed\n");
			exit (1);
		}
			  
#if 1
		printf ("SelectionNotify - returned data %d left %d\n",
				p.nitems, left);
#endif
	  	  
		iReturn = XGetWindowProperty(pDisplay,
									 iWindow,
									 atomLocalProperty,
									 0,
									 left,
									 False,
									 atomCompoundText,
									 &p.encoding,
									 &p.format,
									 &p.nitems,
									 &ulReturnBytesLeft,
									 &p.value);

		if (iReturn != Success || p.nitems <= 0)
		{
			printf ("SelectionNotify - XGetWindowProperty ()2 failed\n");
			printf ("%s required, actual %s\n",
					XGetAtomName(pDisplay, atomCompoundText),
					XGetAtomName(pDisplay, p.encoding));
			//exit (1);
			return;
		}
			  
#if 1
		printf ("SelectionNotify - returned data %d left %d\n",
				p.nitems, ulReturnBytesLeft);
		//printf("%d %d %s, %d\n", p.nitems, p.value, p.value, n);
#endif
		XmbTextPropertyToTextList (pDisplay, &p, &l, &n);
		//printf("%d, %s\n", l, l[0]);
		pszReturnData = malloc((p.nitems+1)*4);
		if (n-- > 0)
		{
			char** lp = l;
			strcpy (pszReturnData, *lp++);
			while (n-- > 0)
			{
				strcat (pszReturnData, *lp++);
			}
		}
		if(l){
			XFreeStringList (l);
		}
		XFree (p.value);
		//printf("Compound Text is '%s'\n", pszReturnData);
		//exit(0);
	}
	else
	{
		printf("Unknown format\n");
		exit(0);
	}
	/* Check Selection is modifiey or not. */
	if(pszPrevClipboardData && strcmp(pszReturnData, pszPrevClipboardData)==0)
	{
#if 1
		printf("Selection Not Modified.\n");
#endif
		return;
	}
	//printf("New Selection is '%s'\n", pszReturnData);
	XFree(pszPrevClipboardData);
	pszPrevClipboardData = malloc(strlen(pszReturnData)+1);
	/* Access the Windows clipboard */
	if (!OpenClipboard (hwnd))
	{
		printf ("OpenClipboard () failed\n");
		PrintWindowsLastError();
		exit (1);
	}
  
	/* Take ownership of the Window clipboard */
	if (!EmptyClipboard ())
	{
		printf ("EmptyClipboard () failed\n");
		PrintWindowsLastError();
		exit (1);
	}
		  
	/* Allocate global memory for the X clipboard data */
	hGlobal = GlobalAlloc (GMEM_MOVEABLE, strlen (pszReturnData) + 1);
		  
	/* Obtain a pointer to the global memory */
	pszGlobalData = GlobalLock (hGlobal);
	if (pszGlobalData == NULL)
	{
		printf ("Could not lock global memory for clipboard transfer\n");
		exit (1);
	}
		  
	/* Copy the returned string into the global memory */
	strcpy (pszGlobalData, pszReturnData);
		  
	/* Free the data returned from XGetWindowProperty */
	XFree (pszReturnData);
	pszReturnData = NULL;   
		  
	/* Release the pointer to the global memory */
	GlobalUnlock (hGlobal);
	pszGlobalData = NULL;
		  
	/* Push the selection data to the Windows clipboard */	  
	if(!SetClipboardData (CF_TEXT, hGlobal)){
		PrintWindowsLastError();
		exit(1);
	}
		  
	/*
	 * NOTE: Do not try to free pszGlobalData, it is owned by
	 * Windows after the call to SetClipboardData ().
	 */
	
	/* Release the clipboard */
	if (!CloseClipboard ())
	{
		printf ("CloseClipboard () failed\n");
		PrintWindowsLastError();
		exit (1);
	}
#if 0
	/* Reassert ownership of the selection */	  
	iReturn = XSetSelectionOwner (pDisplay,
								  event->xselection.selection,
								  iWindow, CurrentTime);
	if (iReturn == BadAtom || iReturn == BadWindow)
	{
		printf ("Could not reassert ownership of selectionn");
		exit (1);
	}
	isOwner = TRUE;
#endif
}
int
OnExit(Display* dsp)
{
	printf("OnExit\n");
	ChangeClipboardChain(hwnd, hwndNextViewer);
	PostQuitMessage(0); 
}
HWND
CreateMessagingWindow ()
{
	WNDCLASS		wc;
	HWND			hwnd;

	/* Setup our window class */
	wc.style = CS_HREDRAW | CS_VREDRAW;
	wc.lpfnWndProc = WindowProc;
	wc.cbClsExtra = 0;
	wc.cbWndExtra = 0;
	wc.hInstance = GetModuleHandle (NULL);
	wc.hIcon = 0;
	wc.hCursor = 0;
	wc.hbrBackground = (HBRUSH) GetStockObject (WHITE_BRUSH);
	wc.lpszMenuName = NULL;
	wc.lpszClassName = WINDOW_CLASS;
	RegisterClass (&wc);

	/* Create the window */
	hwnd = CreateWindowExA (0,			/* Extended styles */
							WINDOW_CLASS,		/* Class name */
							WINDOW_TITLE,		/* Window name */
							WS_OVERLAPPED,	/* Not visible anyway */
							CW_USEDEFAULT,	/* Horizontal position */
							CW_USEDEFAULT,	/* Vertical position */
							CW_USEDEFAULT,	/* Right edge */
							CW_USEDEFAULT,	/* Bottom edge */
							(HWND) NULL,		/* No parent or owner window */
							(HMENU) NULL,		/* No menu */
							GetModuleHandle (NULL),/* Instance handle */
							NULL);		/* ScreenPrivates */
	//ShowWindow(hwnd, SW_SHOW);
	assert (hwnd != NULL);

	return hwnd;
}

void
PrintWindowsLastError()
{
	LPVOID lpMsgBuf;
	FormatMessage(
			FORMAT_MESSAGE_ALLOCATE_BUFFER |
			FORMAT_MESSAGE_FROM_SYSTEM |
			FORMAT_MESSAGE_IGNORE_INSERTS,
			NULL,
			GetLastError(),
			MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),
			(LPTSTR) &lpMsgBuf,
			0,
			NULL
		);
	printf("%s\n", lpMsgBuf);
	LocalFree(lpMsgBuf);
}
